from typing import Callable, List

import numpy as np
import pandas as pd

from sklearn.metrics import auc, precision_recall_curve
from sksurv.metrics import concordance_index_censored

def auprc(y_true: np.ndarray, y_scores: np.ndarray) -> float:
    """
    Calculate the Area Under the Precision-Recall Curve (AUPRC).

    Args:
        y_true (np.ndarray): True binary labels. Shape (n_samples,).
        y_scores (np.ndarray): Predicted scores or probabilities. Shape (n_samples,).

    Returns:
        auprc_score (float): The AUPRC value.
    """
    precision, recall, _ = precision_recall_curve(y_true, y_scores)
    return auc(recall, precision)

def ci_wrapper(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Wrapper for `sksurv.metrics.concordance_index_censored` to ensure compatibility 
    with `bootstrap_metric`.

    Args:
        y_true (np.ndarray): A 2D NumPy array of shape (n_samples, 2), where:
            - `y_true[:, 0]` represents the observed survival times.
            - `y_true[:, 1]` represents the event indicator 
              (1 if the event occurred, 0 if censored).
        y_pred (np.ndarray): A 1D NumPy array of predicted risk scores or 
            survival times. Higher scores typically indicate higher risk.

    Returns:
        concordance_index (float): The concordance index.
    """
    time = y_true[:, 0]
    event = y_true[:, 1]

    return concordance_index_censored(event.astype(bool), time, y_pred)[0]

def bootstrap_metric(
        y_true: np.ndarray,
        y_pred: np.ndarray,
        metric_func: Callable[[np.ndarray, np.ndarray], float],
        nsamples: int = 100
    ) -> List[float]:
    """
    Compute a metric using bootstrapping to estimate its variability.

    Args:
        y_true (np.ndarray): True labels. Shape (n_samples,).
        y_pred (np.ndarray): Predicted values. Shape (n_samples,).
        metric_func (Callable[[np.ndarray, np.ndarray], float]): A function that calculates the metric.
        nsamples (int, optional): The number of bootstrap samples. Defaults to 100.

    Returns:
        bootstrapped_values (List[float]): A list of metric values computed on each bootstrap sample.
    """
    np.random.seed(0)
    values = []

    for _ in range(nsamples):
        idx = np.random.randint(len(y_true), size=len(y_true))
        pred_sample = y_pred[idx]
        y_true_sample = y_true[idx]
        val = metric_func(y_true_sample, pred_sample)
        values.append(val)

    return values

def undummify(df, prefix_sep="_"):
    """
    Undummifies a DataFrame by collapsing dummy/one-hot encoded columns back into their original categorical column.

    Found here: https://stackoverflow.com/a/62085741

    Args:
        df (pandas.DataFrame): The input DataFrame containing dummy/one-hot encoded columns.
        prefix_sep (str, optional): The separator used to distinguish between the prefix (category) and the column name in the dummy columns. 
            Defaults to "_".

    Returns:
        undummified_df (pandas.DataFrame): A new DataFrame with the undummified (reconstructed) categorical columns.
    """
    dummy_cols = {
        item.split(prefix_sep)[0]: (prefix_sep in item) for item in df.columns
    }
    series_list = []
    for col, needs_to_collapse in dummy_cols.items():
        if needs_to_collapse:
            undummified = (
                df.filter(like=col)
                .idxmax(axis=1)
                .apply(lambda x: x.split(prefix_sep, maxsplit=1)[1])
                .rename(col)
            )
            series_list.append(undummified)
        else:
            series_list.append(df[col])
    undummified_df = pd.concat(series_list, axis=1)
    return undummified_df

def process_RADCURE_clinical(df):
    """
    Processes RADCURE clinical data.

    Raw data found here: https://www.cancerimagingarchive.net/collection/radcure/
    """
    df_converted = pd.DataFrame({
        'Study ID': df['patient_id'],
        'survival_time': df['Length FU'],
        'death': df['Status'].apply(lambda x: 1 if x == 'Dead' else 0),
        'age at dx': df['Age'],
        'Sex': df['Sex'],
        'T Stage': df['T'],
        'N Stage': df['N'],
        'Stage': df['Stage'],
        'Dose': df['Dose'],
        'Chemotherapy': df['Chemo'].apply(lambda x: 1 if x != 'none' else 0),
        'HPV Combined': df['HPV'].apply(lambda x: 1 if isinstance(x, str) and 'positive' in x.lower() else None),
        'Smoking Status': df['Smoking Status'],
        'Disease Site': df['Ds Site'].str.lower()
    })
    
    return df_converted


import json
import os
import pickle
import shutil
from pathlib import Path
from typing import Tuple

import pandas as pd
from autogluon.tabular import TabularPredictor
from sklearn.model_selection import KFold, train_test_split
from sksurv.ensemble import GradientBoostingSurvivalAnalysis, RandomSurvivalForest
from sksurv.linear_model import CoxnetSurvivalAnalysis
from sksurv.metrics import concordance_index_censored
from sksurv.svm import FastSurvivalSVM
from sksurv.util import Surv

from ._leaderboard import aggregate_folds, format_leaderboard
from .survival import train_deepsurv, train_mtlr

def train_autogluon_with_cv(
        data_train: pd.DataFrame, data_test: pd.DataFrame,
        target_variable: str, task: str,
        output_dir: Path, extra_metrics: list,
        eval_metric: str='accuracy', num_folds: int=5, **kwargs: dict):
    """
    Trains a TabularPredictor using manual cross-validation without bagging and consolidates the leaderboards.

    Parameters
    ----------
    - data_train (DataFrame): Combined training data (features + target).
    - data_test (DataFrame): Combined test data (features + target).
    - target_variable (str): Name of the target column.
    - task (str): Problem type (e.g., 'binary', 'multiclass', 'regression').
    - output_dir (Path): Directory to save model files.
    - eval_metric (str): Evaluation metric to optimize (default: 'accuracy').
    - num_folds (int): Number of cross-validation folds (default: 5).
    - kwargs (dict): Additional arguments to pass to TabularPredictor's fit method.

    Returns
    -------
    - predictors: A list of trained predictors (one per fold).
    - final_leaderboard: A single DataFrame containing all models across folds.
    - best_fold: The index of best training fold
    - X_val: The validation features 
    - y_val: The validation target
    """
    kf = KFold(n_splits=num_folds, shuffle=True, random_state=42)

    predictors, cv_scores, val_indices = [], [], []
    train_leaderboards, val_leaderboards, test_leaderboards = [], [], []

    for fold, (train_idx, val_idx) in enumerate(kf.split(data_train)):
        print(f"Training fold {fold + 1}/{num_folds}...")

        train_data, val_data = data_train.iloc[train_idx], data_train.iloc[val_idx]
        val_indices.append(val_idx)

        predictor = TabularPredictor(
            label=target_variable, problem_type=task, eval_metric=eval_metric,
            path=os.path.join(output_dir, f'autogluon_models_fold_{fold + 1}'),
            verbosity=0, log_to_file=False,
        ).fit(
            train_data,
            tuning_data=val_data,
            use_bag_holdout=True,
            **kwargs
        )

        score = predictor.evaluate(val_data)[eval_metric]
        print(f"Fold {fold + 1} score: {score}")

        predictors.append(predictor)
        cv_scores.append(score)

        train_leaderboards.append(predictor.leaderboard(train_data, extra_metrics=extra_metrics))
        val_leaderboards.append(predictor.leaderboard(val_data, extra_metrics=extra_metrics))
        test_leaderboards.append(predictor.leaderboard(data_test, extra_metrics=extra_metrics))

    train_leaderboard = aggregate_folds(pd.concat(train_leaderboards, ignore_index=True), extra_metrics)
    val_leaderboard = aggregate_folds(pd.concat(val_leaderboards, ignore_index=True), extra_metrics)
    test_leaderboard = aggregate_folds(pd.concat(test_leaderboards, ignore_index=True), extra_metrics)

    final_leaderboard = pd.merge(
        pd.merge(
            format_leaderboard(train_leaderboard, extra_metrics, 'score_train'),
            format_leaderboard(val_leaderboard, extra_metrics, 'score_val'),
            on='model'
        ),
        format_leaderboard(test_leaderboard, extra_metrics, 'score_test'),
        on='model'
    )

    best_fold = cv_scores.index(max(cv_scores))
    val_indices_best = val_indices[best_fold]
    X_val = data_train.iloc[val_indices_best].drop(columns=target_variable)
    y_val = data_train.iloc[val_indices_best][target_variable]

    shutil.copytree(os.path.join(output_dir, f'autogluon_models_fold_{best_fold + 1}'),
                    os.path.join(output_dir, 'autogluon_models_best_fold'), dirs_exist_ok=True)

    return predictors, final_leaderboard, best_fold, X_val, y_val

def train_survival_models(
        X_train: pd.DataFrame, 
        y_train: pd.DataFrame,
        X_test: pd.DataFrame, 
        y_test: pd.DataFrame, 
        output_dir: Path
    ):
    """Train both deep and traditional survival models, consolidate fitted models and C-index scores."""
    (output_dir / 'survival_models').mkdir(exist_ok=True, parents=True)

    fitted_models = {}
    cindex_scores = {}

    # Deep Models

    data_train, data_val = train_test_split(pd.concat([X_train, y_train], axis=1), test_size=0.1, stratify=y_train['event'], random_state=42)

    fitted_models['MTLR'], cindex_scores['MTLR'] = train_mtlr(
        data_train,
        data_val,
        pd.concat([X_test, y_test], axis=1),
        output_dir / 'survival_models')
    fitted_models['DeepSurv'], cindex_scores['DeepSurv'] = train_deepsurv(
        data_train,
        data_val,
        pd.concat([X_test, y_test], axis=1),
        output_dir / 'survival_models')

    # Basic Models
    y_train = Surv.from_dataframe('event', 'time', y_train)
    y_test = Surv.from_dataframe('event', 'time', y_test)

    models = {
        "CoxPH": CoxnetSurvivalAnalysis(fit_baseline_model=True),
        "GradientBoosting": GradientBoostingSurvivalAnalysis(),
        "RandomForest": RandomSurvivalForest(n_estimators=100, random_state=42),
        "SVM": FastSurvivalSVM(max_iter=1000, tol=1e-5, random_state=42),
    }

    for name, model in models.items():
        print(f"Training {name}...")
        model.fit(X_train, y_train)
        fitted_models[name] = model

        model_path = output_dir / 'survival_models' / f"{name}.pkl"
        with open(model_path, "wb") as f:
            pickle.dump(model, f)

        predictions = model.predict(X_test)
        cindex_scores[name] = concordance_index_censored(
            y_test["event"], y_test["time"], predictions
        )[0]

    print("\nConsolidated C-index Scores:")
    for model_name, cindex in cindex_scores.items():
        print(f"{model_name}: {cindex:.4f}")

    # For later saving to yaml
    cindex_scores = {key: float(value) for key, value in cindex_scores.items()}

    return fitted_models, cindex_scores, data_train, data_val



# ------------------------------------------------------------------------------
# Code adapted from:
# https://github.com/czifan/DeepSurv.pytorch
# ------------------------------------------------------------------------------

import numpy as np

import torch
from torch.utils.data import Dataset

from lifelines.utils import concordance_index

def calculate_c_index(risk_pred, y, e):
    ''' Performs calculating c-index

    :param risk_pred: (np.ndarray or torch.Tensor) model prediction
    :param y: (np.ndarray or torch.Tensor) the times of event e
    :param e: (np.ndarray or torch.Tensor) flag that records whether the event occurs
    :return c_index: the c_index is calculated by (risk_pred, y, e)
    '''
    if not isinstance(y, np.ndarray):
        y = y.detach().cpu().numpy()
    if not isinstance(risk_pred, np.ndarray):
        risk_pred = risk_pred.detach().cpu().numpy()
    if not isinstance(e, np.ndarray):
        e = e.detach().cpu().numpy()
    return concordance_index(y, risk_pred, e)

class SurvivalDataset(Dataset):
    def __init__(self, data):

        self.X = data.drop(columns=['time', 'event']).to_numpy(dtype=np.float32)
        self.e = data['event'].to_numpy(dtype=np.float32).reshape(-1, 1)
        self.y = data['time'].to_numpy(dtype=np.float32).reshape(-1, 1)

        self._normalize()

    def _normalize(self):
        '''Performs normalizing X data.'''
        range_values = self.X.max(axis=0) - self.X.min(axis=0)
        range_values[range_values == 0] = 1  # Prevent division by zero
        self.X = (self.X - self.X.min(axis=0)) / range_values

    def __getitem__(self, item):
        ''' Performs constructing torch.Tensor object'''
        # gets data with index of item
        X_item = self.X[item] # (m)
        e_item = self.e[item] # (1)
        y_item = self.y[item] # (1)
        # constructs torch.Tensor object
        X_tensor = torch.from_numpy(X_item)
        e_tensor = torch.from_numpy(e_item)
        y_tensor = torch.from_numpy(y_item)
        return X_tensor, y_tensor, e_tensor

    def __len__(self):
        return self.X.shape[0]


import logging
from pathlib import Path

import pandas as pd
import numpy as np

import torch
from torch.utils.data import DataLoader, TensorDataset

from .mtlr import mtlr_risk, LitMTLR
from .utils import encode_survival, make_time_bins, normalize
from lifelines.utils import concordance_index

import lightning.pytorch as pl
import optuna
from optuna.integration import PyTorchLightningPruningCallback

# Suppress PyTorch Lightning logging
logging.getLogger("lightning.pytorch.utilities.rank_zero").setLevel(logging.FATAL)

def calculate_c_index(risk_pred, y, e):
    ''' Performs calculating c-index

    :param risk_pred: (np.ndarray or torch.Tensor) model prediction
    :param y: (np.ndarray or torch.Tensor) the times of event e
    :param e: (np.ndarray or torch.Tensor) flag that records whether the event occurs
    :return c_index: the c_index is calculated by (risk_pred, y, e)
    '''
    if not isinstance(y, np.ndarray):
        y = y.detach().cpu().numpy()
    if not isinstance(risk_pred, np.ndarray):
        risk_pred = risk_pred.detach().cpu().numpy()
    if not isinstance(e, np.ndarray):
        e = e.detach().cpu().numpy()
    return concordance_index(y, risk_pred, e)

def create_dataloader(data: pd.DataFrame, time_bins: torch.Tensor):
    X = torch.tensor(data.drop(["time", "event"], axis=1).values, dtype=torch.float)
    y = encode_survival(data["time"].values, data["event"].values, time_bins)

    return DataLoader(TensorDataset(X, y), batch_size=len(data), shuffle=True)

def train_mtlr(data_train: pd.DataFrame, data_val:pd.DataFrame, data_test: pd.DataFrame, output_dir: Path):

    in_channel = len(data_train.columns) - 2  # -2 to exclude "time" and "event"
    time_bins = make_time_bins(data_train["time"].values, event=data_train["event"].values)

    skip_cols = [
        col for col in data_train.columns if (set(data_train[col].unique()).issubset({0, 1}) or (col in ['time', 'event']))
    ]
    data_train, mean, std = normalize(data_train, skip_cols=skip_cols)
    data_val, _, _ = normalize(data_val, mean=mean, std=std, skip_cols=skip_cols)
    data_test, _, _ = normalize(data_test, mean=mean, std=std, skip_cols=skip_cols)

    train_loader = create_dataloader(data_train, time_bins)
    val_loader = create_dataloader(data_val, time_bins)

    def objective(trial: optuna.trial.Trial) -> float:
        C1 = trial.suggest_categorical("C1", [c1 for c1 in np.logspace(-2, 3, 6)])
        dropout = trial.suggest_float("dropout", 0.2, 0.5)
        dims = trial.suggest_categorical("dims", [[2**n, 2**n] for n in range(4, 10)])

        model = LitMTLR(in_channel=in_channel, num_time_bins=len(time_bins), dims=dims, dropout=dropout, C1=C1)

        trainer = pl.Trainer(
            enable_checkpointing=False,
            enable_progress_bar=False,
            enable_model_summary=False,
            max_epochs=10,
            accelerator="auto",
            callbacks=[PyTorchLightningPruningCallback(trial, monitor="valid_loss")],
        )
        trainer.fit(model, train_loader, val_loader)

        return trainer.callback_metrics["valid_loss"].item()

    pruner = optuna.pruners.MedianPruner()

    optuna.logging.set_verbosity(optuna.logging.WARNING)
    study = optuna.create_study(direction="minimize", pruner=pruner)

    print('Training MTLR...')
    study.optimize(objective, n_trials=300, timeout=600)

    print("  Best trial:")
    trial = study.best_trial
    print("    Params: ")
    for key, value in trial.params.items():
        print("      {}: {}".format(key, value))

    # Train the final model with the best parameters
    model = LitMTLR(in_channel=in_channel, num_time_bins=len(time_bins), mean=mean, std=std, **trial.params)

    trainer = pl.Trainer(
        default_root_dir=output_dir,
        enable_checkpointing=False,
        enable_progress_bar=False,
        enable_model_summary=False,
        max_epochs=500,
        accelerator="auto",
        callbacks=[
            pl.callbacks.EarlyStopping(monitor="valid_loss", mode="min")
        ]
    )
    trainer.fit(model, train_loader, val_loader)
    trainer.save_checkpoint(output_dir / 'MTLR.ckpt')

    # Evaluate on the test set
    model.eval()
    with torch.no_grad():
        y_pred = model(torch.tensor(data_test.drop(["time", "event"], axis=1).values, dtype=torch.float))
        risk_pred = mtlr_risk(y_pred)
    c_index = calculate_c_index(-risk_pred, data_test['time'].values, data_test["event"].values)

    return model, c_index



from jarvais import TrainerSupervised
import pandas as pd

df = pd.read_csv('data/a2rData_train_ED.csv')

trainer = TrainerSupervised(task='binary', output_dir='BRUHTEST')

trainer.run(data=df, target_variable='target_ED_visit', k_folds=2)


import pandas as pd

from typing import Callable

def feature_engineering_imaging(data: pd.DataFrame):
    """
    Function to perform feature engineering on imaging data using PyRadiomics.
    
    Parameters:
    - data (DataFrame): The input imaging data.
    
    Returns:
    - radiomics_features (DataFrame): DataFrame containing radiomics features.
    """
    
    raise NotImplementedError


symp_cols = [
    'esas_pain',
    'esas_tiredness',
    'esas_nausea',
    'esas_depression',
    'esas_anxiety',
    'esas_drowsiness',
    'esas_appetite',
    'esas_well_being',
    'esas_shortness_of_breath',
    'patient_ecog',
]
lab_cols = [
    'alanine_aminotransferase',
    'albumin',
    'alkaline_phosphatase',
    'aspartate_aminotransferase',
    #'basophil',
    'bicarbonate',
    'chloride',
    'creatinine',
    'eosinophil',
    'glucose',
    'hematocrit',
    'hemoglobin',
    'lactate_dehydrogenase',
    'lymphocyte',
    'magnesium',
    'mean_corpuscular_hemoglobin',
    'mean_corpuscular_hemoglobin_concentration',
    'mean_corpuscular_volume',
    'mean_platelet_volume',
    'monocyte',
    'neutrophil',
    'phosphate',
    'platelet',
    'potassium',
    'red_blood_cell',
    'red_cell_distribution_width',
    'sodium',
    'total_bilirubin',
    'white_blood_cell',
]
symp_change_cols = [f'{col}_change' for col in symp_cols]
lab_change_cols = [f'{col}_change' for col in lab_cols]


def feature_engineering_clinical(data: pd.DataFrame, cfg): #, function: Callable
    """
    Function to perform feature engineering on clinical data using custom functions.
    
    Parameters:
    - data (DataFrame): The input clinical data.
    - function_file (str): The path to the file containing customized feature engineering functions.
    
    Returns:
    - clinical_features (DataFrame): DataFrame containing clinical features.
    """
    import numpy as np
    from functools import partial
    # import yaml
    # from constants import lab_cols, lab_change_cols, symp_cols, symp_change_cols

    # with open('config.yaml') as file:
    #     cfg = yaml.safe_load(file)
        
    # align_on = cfg['align_on'] 
    main_date_col = cfg['main_date_col'] 

    def get_change_since_prev_session(data: pd.DataFrame) -> pd.DataFrame:
        """Get change since last session"""
        cols = symp_cols + lab_cols
        change_cols = symp_change_cols + lab_change_cols
        result = []
        for mrn, group in data.groupby('mrn'):
            group[cols] = group[cols].apply(pd.to_numeric) # convert all columns of DataFrame # pd.to_numeric(s, errors='coerce')
            change = group[cols] - group[cols].shift()
            result.append(change.reset_index().to_numpy())
        result = np.concatenate(result)

        result = pd.DataFrame(result, columns=['index']+change_cols).set_index('index')
        result.index = result.index.astype(int)
        data = pd.concat([data, result], axis=1)

        return data

    def get_missingness_features(data: pd.DataFrame) -> pd.DataFrame:
        
        target_cols = symp_cols + lab_cols + lab_change_cols + symp_change_cols
        
        for col in target_cols: data[f'{col}_is_missing'] = data[col].isnull()
            
        return data

    ###############################################################################
    # Time
    ###############################################################################
    def get_visit_month_feature(data, col: str = 'treatment_date'):
        # convert to cyclical features
        month = data[col].dt.month - 1
        data['visit_month_sin'] = np.sin(2*np.pi*month/12)
        data['visit_month_cos'] = np.cos(2*np.pi*month/12)
        return data

    def get_days_since_last_event(data, main_date_col: str = 'treatment_date', event_date_col: str = 'treatment_date'):
        if main_date_col == event_date_col:
            return (data[main_date_col] - data[event_date_col].shift()).dt.days
        else:
            return (data[main_date_col] - data[event_date_col]).dt.days

    def get_years_diff(data, col1: str, col2: str):
        return data[col1].dt.year - data[col2].dt.year

    ###############################################################################
    # Treatment
    ###############################################################################
    def get_line_of_therapy(data):
        # identify line of therapy (the nth different palliative intent treatment taken)
        # NOTE: all other intent treatment are given line of therapy of 0. Usually (not always but oh well) once the first
        # palliative treatment appears, the rest of the treatments remain palliative
        new_regimen = (data['first_treatment_date'] != data['first_treatment_date'].shift())
        palliative_intent = data['intent'] == 'PALLIATIVE'
        return (new_regimen & palliative_intent).cumsum()


    ###############################################################################
    # Drug dosages
    ###############################################################################
    def get_perc_ideal_dose_given(data, drug_to_dose_formula_map: dict[str, str]):
        """Convert given dose as a percentage of ideal (recommended) dose

        data must have weight, body surface area, age, female, and creatinine columns along with the dosage columns
        """
        result = {}
        for drug, dose_formula in drug_to_dose_formula_map.items():
            dose_col = f'drug_{drug}_given_dose'
            if dose_col not in data.columns: continue
            ideal_dose = get_ideal_dose(data, drug, dose_formula)
            perc_ideal_dose_given = data[dose_col] / ideal_dose # NOTE: 0/0 = np.nan, x/0 = np.inf
            perc_ideal_dose_given = perc_ideal_dose_given
            result[drug] = perc_ideal_dose_given
        result = pd.DataFrame(result)
        result.columns = '%_ideal_dose_given_' + result.columns
        return result

    def get_ideal_dose(data, drug: str, dose_formula: str):
        col = f'drug_{drug}_regimen_dose'
        carboplatin_dose_formula = ('min(regimen_dose * 150, regimen_dose * (((140-age[yrs]) * weight [kg] * 1.23 * '
                                    '(0.85 if female) / creatinine [umol/L]) + 25))')
        if dose_formula == 'regimen_dose': 
            return data[col]
        
        elif dose_formula == 'regimen_dose * bsa': 
            return data[col] * data['body_surface_area']
        
        elif dose_formula == 'regimen_dose * weight': 
            return data[col] * data['weight']
        
        elif dose_formula == carboplatin_dose_formula:
            return pd.concat([data[col] * 150, data[col] * (get_creatinine_clearance(data) + 25)], axis=1).min(axis=1)
        
        else:
            raise ValueError(f'Ideal dose formula {dose_formula} not supported')

    ###############################################################################
    # Special Formulas
    ###############################################################################
    def get_creatinine_clearance(data):
        return (140 - data['age']) * data['weight'] * 1.23 * data['female'].replace({True: 0.85, False: 1}) / data['creatinine']
    
    
    ######### Process feature engineering ##########
    data['treatment_date'] = pd.to_datetime(data['treatment_date'], errors='coerce')
    data['first_treatment_date'] = pd.to_datetime(data['first_treatment_date'], errors='coerce')
    
    data = get_change_since_prev_session(data)
    data = get_missingness_features(data)
    data = get_visit_month_feature(data, col=main_date_col)
    data['line_of_therapy'] = data.groupby('mrn', group_keys=False).apply(get_line_of_therapy)
    data['days_since_starting_treatment'] = (data[main_date_col] - data['first_treatment_date']).dt.days
    get_days_since_last_treatment = partial(
        get_days_since_last_event, main_date_col=main_date_col, event_date_col='treatment_date'
    )
    data['days_since_last_treatment'] = data.groupby('mrn', group_keys=False).apply(get_days_since_last_treatment)

    return data

    # raise NotImplementedError
    
# ###############################################################################
# # Engineering Features
# ###############################################################################
# def custom_clinical_features(df: pd.DataFrame):
#     """
#     Perform customized feature engineering on the given DataFrame.

#     Parameters:
#     df (pd.DataFrame): The input DataFrame.

#     Returns:
#     pd.DataFrame: The transformed DataFrame with engineered features.
#     """
#     import numpy as np
#     from functools import partial
#     import yaml
#     from constants import lab_cols, lab_change_cols, symp_cols, symp_change_cols

#     with open('config.yaml') as file:
#         cfg = yaml.safe_load(file)
        
#     # align_on = cfg['align_on'] 
#     main_date_col = cfg['main_date_col'] 

#     def get_change_since_prev_session(df: pd.DataFrame) -> pd.DataFrame:
#         """Get change since last session"""
#         cols = symp_cols + lab_cols
#         change_cols = symp_change_cols + lab_change_cols
#         result = []
#         for mrn, group in df.groupby('mrn'):
#             group[cols] = group[cols].apply(pd.to_numeric) # convert all columns of DataFrame # pd.to_numeric(s, errors='coerce')
#             change = group[cols] - group[cols].shift()
#             result.append(change.reset_index().to_numpy())
#         result = np.concatenate(result)

#         result = pd.DataFrame(result, columns=['index']+change_cols).set_index('index')
#         result.index = result.index.astype(int)
#         df = pd.concat([df, result], axis=1)

#         return df

#     def get_missingness_features(df: pd.DataFrame) -> pd.DataFrame:
        
#         target_cols = symp_cols + lab_cols + lab_change_cols + symp_change_cols
        
#         for col in target_cols: df[f'{col}_is_missing'] = df[col].isnull()
            
#         return df

#     ###############################################################################
#     # Time
#     ###############################################################################
#     def get_visit_month_feature(df, col: str = 'treatment_date'):
#         # convert to cyclical features
#         month = df[col].dt.month - 1
#         df['visit_month_sin'] = np.sin(2*np.pi*month/12)
#         df['visit_month_cos'] = np.cos(2*np.pi*month/12)
#         return df

#     def get_days_since_last_event(df, main_date_col: str = 'treatment_date', event_date_col: str = 'treatment_date'):
#         if main_date_col == event_date_col:
#             return (df[main_date_col] - df[event_date_col].shift()).dt.days
#         else:
#             return (df[main_date_col] - df[event_date_col]).dt.days

#     def get_years_diff(df, col1: str, col2: str):
#         return df[col1].dt.year - df[col2].dt.year

#     ###############################################################################
#     # Treatment
#     ###############################################################################
#     def get_line_of_therapy(df):
#         # identify line of therapy (the nth different palliative intent treatment taken)
#         # NOTE: all other intent treatment are given line of therapy of 0. Usually (not always but oh well) once the first
#         # palliative treatment appears, the rest of the treatments remain palliative
#         new_regimen = (df['first_treatment_date'] != df['first_treatment_date'].shift())
#         palliative_intent = df['intent'] == 'PALLIATIVE'
#         return (new_regimen & palliative_intent).cumsum()


#     ###############################################################################
#     # Drug dosages
#     ###############################################################################
#     def get_perc_ideal_dose_given(df, drug_to_dose_formula_map: dict[str, str]):
#         """Convert given dose as a percentage of ideal (recommended) dose

#         df must have weight, body surface area, age, female, and creatinine columns along with the dosage columns
#         """
#         result = {}
#         for drug, dose_formula in drug_to_dose_formula_map.items():
#             dose_col = f'drug_{drug}_given_dose'
#             if dose_col not in df.columns: continue
#             ideal_dose = get_ideal_dose(df, drug, dose_formula)
#             perc_ideal_dose_given = df[dose_col] / ideal_dose # NOTE: 0/0 = np.nan, x/0 = np.inf
#             perc_ideal_dose_given = perc_ideal_dose_given
#             result[drug] = perc_ideal_dose_given
#         result = pd.DataFrame(result)
#         result.columns = '%_ideal_dose_given_' + result.columns
#         return result

#     def get_ideal_dose(df, drug: str, dose_formula: str):
#         col = f'drug_{drug}_regimen_dose'
#         carboplatin_dose_formula = ('min(regimen_dose * 150, regimen_dose * (((140-age[yrs]) * weight [kg] * 1.23 * '
#                                     '(0.85 if female) / creatinine [umol/L]) + 25))')
#         if dose_formula == 'regimen_dose': 
#             return df[col]
        
#         elif dose_formula == 'regimen_dose * bsa': 
#             return df[col] * df['body_surface_area']
        
#         elif dose_formula == 'regimen_dose * weight': 
#             return df[col] * df['weight']
        
#         elif dose_formula == carboplatin_dose_formula:
#             return pd.concat([df[col] * 150, df[col] * (get_creatinine_clearance(df) + 25)], axis=1).min(axis=1)
        
#         else:
#             raise ValueError(f'Ideal dose formula {dose_formula} not supported')

#     ###############################################################################
#     # Special Formulas
#     ###############################################################################
#     def get_creatinine_clearance(df):
#         return (140 - df['age']) * df['weight'] * 1.23 * df['female'].replace({True: 0.85, False: 1}) / df['creatinine']
    

#     df = get_visit_month_feature(df, col=main_date_col)
#     df['line_of_therapy'] = df.groupby('mrn', group_keys=False).apply(get_line_of_therapy)
#     df['days_since_starting_treatment'] = (df[main_date_col] - df['first_treatment_date']).dt.days
#     get_days_since_last_treatment = partial(
#         get_days_since_last_event, main_date_col=main_date_col, event_date_col='treatment_date'
#     )
#     df['days_since_last_treatment'] = df.groupby('mrn', group_keys=False).apply(get_days_since_last_treatment)

#     return df


from pathlib import Path
from typing import List

import pandas as pd
from sklearn.inspection import permutation_importance
from sksurv.util import Surv

from ..utils.pdf import generate_explainer_report_pdf
from ..utils.functional import undummify
from ..utils.plot import (
    plot_classification_diagnostics,
    plot_feature_importance,
    plot_regression_diagnostics,
    plot_shap_values,
    plot_violin_of_bootstrapped_metrics,
)
from .bias import BiasExplainer, infer_sensitive_features

class Explainer():
    """
    A class to generate diagnostic plots and reports for models trained using TrainerSupervised.

    Attributes:
        trainer (TrainerSupervised): The TrainerSupervised object containing the trained model.
        predictor (object): The AutoGluon predictor object used for inference.
        X_train (pd.DataFrame): The training dataset used to train the model.
        X_test (pd.DataFrame): The test dataset for evaluating the model.
        y_test (pd.DataFrame): The true target values for the test dataset.
        output_dir (Path): The directory where plots, reports, and outputs are saved.
        sensitive_features (list, optional): List of features considered sensitive for bias auditing.
    """
    def __init__(
            self,
            trainer,
            X_train: pd.DataFrame,
            X_test: pd.DataFrame,
            y_test: pd.DataFrame,
            output_dir: str | Path | None = None,
            sensitive_features: list | None = None,
        ) -> None:

        self.trainer = trainer
        self.predictor = trainer.predictor
        self.X_train = X_train
        self.X_test = X_test
        self.y_test = y_test
        self.sensitive_features = sensitive_features

        self.output_dir = Path.cwd() if output_dir is None else Path(output_dir)
        self.output_dir.mkdir(parents=True, exist_ok=True)
        (self.output_dir / 'figures').mkdir(parents=True, exist_ok=True)

    def run(self) -> None:
        """Generate diagnostic plots and reports for the trained model."""

        self._run_bias_audit()

        plot_violin_of_bootstrapped_metrics(
            self.trainer,
            self.X_test,
            self.y_test,
            self.trainer.X_val,
            self.trainer.y_val,
            self.X_train,
            self.trainer.y_train,
            output_dir=self.output_dir / 'figures'
        )            

        if self.trainer.task in ['binary', 'multiclass']:
            plot_classification_diagnostics(
                self.y_test,
                self.predictor.predict_proba(self.X_test).iloc[:, 1],
                self.trainer.y_val,
                self.predictor.predict_proba(self.trainer.X_val).iloc[:, 1],
                self.trainer.y_train,
                self.predictor.predict_proba(self.X_train).iloc[:, 1],
                output_dir=self.output_dir / 'figures'
            )
            plot_shap_values(
                self.predictor,
                self.X_train,
                self.X_test,
                output_dir=self.output_dir / 'figures'
            )

        elif self.trainer.task == 'regression':
            plot_regression_diagnostics(
                self.y_test,
                self.predictor.predict(self.X_test, as_pandas=False),
                output_dir=self.output_dir / 'figures'
            )

        # Plot feature importance
        if self.trainer.task == 'survival': # NEEDS TO BE UPDATED
            model = self.trainer.predictors['CoxPH']
            result = permutation_importance(model, self.X_test,
                                            Surv.from_dataframe('event', 'time', self.y_test),
                                            n_repeats=15)

            importance_df = pd.DataFrame(
                {
                    "importance": result["importances_mean"],
                    "stddev": result["importances_std"],
                },
                index=self.X_test.columns,
            ).sort_values(by="importance", ascending=False)
            model_name = 'CoxPH'
        else:
            importance_df = self.predictor.feature_importance(
                pd.concat([self.X_test, self.y_test], axis=1))
            model_name = self.predictor.model_best

        plot_feature_importance(importance_df, self.output_dir / 'figures', model_name)
        generate_explainer_report_pdf(self.trainer.task, self.output_dir)

    def _run_bias_audit(self) -> List[pd.DataFrame]:

        bias_output_dir = self.output_dir / 'bias'
        bias_output_dir.mkdir(parents=True, exist_ok=True)

        if self.sensitive_features is None:
            if self.trainer.task == 'survival': # Data needs to be not be one hot encoded
                self.sensitive_features = infer_sensitive_features(undummify(self.X_test, prefix_sep='|'))
            else:
                self.sensitive_features = infer_sensitive_features(self.X_test)
        
        y_pred = None if self.trainer.task == 'survival' else pd.Series(self.trainer.infer(self.X_test) )
        metrics = ['mean_prediction'] if self.trainer.task == 'regression' else ['mean_prediction', 'false_positive_rate'] 

        bias = BiasExplainer(
            self.y_test, 
            y_pred, 
            self.sensitive_features,
            self.trainer.task, 
            bias_output_dir,
            metrics
        )
        bias.run(relative=True)

    @classmethod
    def from_trainer(cls, trainer, **kwargs):
        """Create Explainer object from TrainerSupervised object."""
        return cls(trainer, trainer.X_train, trainer.X_test, trainer.y_test, trainer.output_dir, **kwargs)



from typing import List, Tuple, Callable
from pathlib import Path
from itertools import combinations

import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

import pandas as pd
from tabulate import tabulate

import numpy as np
from lifelines import CoxPHFitter
import statsmodels.api as sm
from statsmodels.stats.multitest import multipletests

import fairlearn as fm

from .utils import _get_metric

def subgroup_analysis_fairlearn(
        y_true: pd.Series,
        y_pred: pd.Series,
        sensitive_feature: pd.Series, 
        metrics: List[str] = ['mean_prediction', 'false_positive_rate', 'true_positive_rate'], 
        fairness_threshold: float = 1.2, 
        relative: bool = False,
        **kwargs: dict
    ) -> pd.DataFrame:
    """
    Computes Fairlearn fairness metrics and returns them in a structured DataFrame.

    Args:
        y_true (pd.Series): Ground truth labels.
        y_pred (pd.Series): Predicted labels or scores.
        sensitive_feature (pd.Series): The feature to analyze fairness for.
        metrics (List[str], optional): List of Fairlearn metrics to compute. Defaults to common fairness metrics.
        fairness_threshold (float, optional): Threshold for relative fairness comparisons. Defaults to 1.2.
        relative (bool, optional): Whether to compute relative fairness metrics. Defaults to False.
        **kwargs (Dict): Additional parameters for `MetricFrame`.

    Returns:
        pd.DataFrame: A DataFrame with computed metrics, optionally with relative fairness values.
    """
    mapper = {
        "mean_prediction": "Demographic Parity",
        "false_positive_rate": "(FPR) Equalized Odds",
        "true_positive_rate": "(TPR) Equalized Odds or Equal Opportunity"
    }

    _metrics = {metric: _get_metric(metric, sensitive_features=sensitive_feature) for metric in metrics}
    metric_frame = fm.MetricFrame(
        metrics=_metrics, 
        y_true=y_true, 
        y_pred=y_pred, 
        sensitive_features=sensitive_feature, 
        **kwargs
    )

    result = pd.DataFrame(metric_frame.by_group.T, index=_metrics.keys())
    result = result.rename(columns=mapper)

    if relative:
        largest_feature = sensitive_feature.mode().iloc[0]
        results_relative = result.T / result[largest_feature]
        results_relative = results_relative.applymap(
            lambda x: f"{x:.3f} ✅" if x <= fairness_threshold or 1/x <= fairness_threshold 
            else f"{x:.3f} ❌")
        result = pd.concat([result, results_relative.T.rename(index=lambda x: f"Relative {x}")])
    
    return result

def subgroup_analysis_OLS(
        sensitive_features: pd.DataFrame,
        endpoint: pd.Series,
        output_dir: Path | str,
        print_report: bool = True, 
        save_report: bool = True,
    ) -> Tuple[sm.OLS, float]:
    """
    Performs Ordinary Least Squares (OLS) regression to analyze the relationship between 
    `sensitive_features` and `endpoint`.

    This function fits an OLS model using one-hot encoded `sensitive_features` to predict `endpoint`. 
    If the overall model p-value is < 0.05, a detailed report is printed and optionally saved.

    Args:
        sensitive_features (pd.DataFrame): DataFrame containing categorical sensitive features.
        endpoint (pd.Series): Series representing the target variable (bias metric).
        output_dir (Path | str): Directory to save the regression report.
        print_report (bool, optional): Whether to print the report. Defaults to True.
        save_report (bool, optional): Whether to save the report. Defaults to True.

    Returns:
        Tuple[sm.OLS, float] The fitted OLS model and the model's F-statistic p-value.
    """
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    feature_names = sensitive_features.columns

    one_hot_encoded = pd.get_dummies(
        sensitive_features[feature_names], 
    )
    X_columns = one_hot_encoded.columns

    X = one_hot_encoded.values  
    y = endpoint.values  

    X = sm.add_constant(X.astype(float), has_constant='add')
    model = sm.OLS(y, X).fit()

    p_value = model.f_pvalue

    if p_value < 0.05 and (print_report or save_report):
        output = []

        print(f"⚠️  **Possible Bias Detected in {feature_names}** ⚠️\n")
        output.append(f"=== Subgroup Analysis for '{feature_names}' Using OLS Regression ===\n")

        output.append("Model Statistics:")
        output.append(f"    R-squared:                  {model.rsquared:.3f}")
        output.append(f"    F-statistic:                {model.fvalue:.3f}")
        output.append(f"    F-statistic p-value:        {p_value:.4f}")
        output.append(f"    AIC:                        {model.aic:.2f}")
        output.append(f"    Log-Likelihood:             {model.llf:.2f}")

        summary_df = pd.DataFrame({
            'Feature': ['const'] + X_columns.tolist(),     # Predictor names (includes 'const' if added)
            'Coefficient': model.params,    # Coefficients
            'Standard Error': model.bse     # Standard Errors
        })
        table_output = tabulate(summary_df, headers='keys', tablefmt='fancy_grid', showindex=False, floatfmt=".3f")
        output.append("Model Coefficients:")
        output.append('\n'.join(['    ' + line for line in table_output.split('\n')]))

        output_text = '\n'.join(output)

        if print_report:
            print(output_text)
        
        if save_report:
            with open(output_dir / f'{'_'.join(feature_names)}_OLS_summary.txt', 'w') as f:
                f.write(output_text)

    return model, p_value
    
def subgroup_analysis_CoxPH(
        sensitive_features: pd.DataFrame,
        endpoint: pd.DataFrame,
        output_dir: Path | str,
        print_report: bool = True, 
        save_report: bool = True,
    ) -> Tuple[CoxPHFitter, float]:
    """
    Fits a Cox Proportional Hazards (CoxPH) model using `sensitive_features` and survival data.

    This function models the hazard ratio for different sensitive feature categories 
    using a CoxPH model. If the likelihood ratio test p-value is < 0.05, a detailed 
    report is printed and optionally saved.

    Args:
        sensitive_features (pd.DataFrame): DataFrame containing categorical sensitive features.
        endpoint (pd.DataFrame): DataFrame with 'time' (duration) and 'event' (censoring status).
        output_dir (Path | str): Directory to save the regression report.
        print_report (bool, optional): Whether to print the report. Defaults to True.
        save_report (bool, optional): Whether to save the report. Defaults to True.

    Raises:
        AssertionError: If 'time' or 'event' columns are missing in `endpoint`.

    Returns:
        Tuple[CoxPHFitter, float] The fitted Cox model and the log-likelihood ratio test p-value.
    """
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    assert {'time', 'event'}.issubset(set(endpoint.columns)), \
    "Error: 'endpoint' DataFrame must contain 'time' and 'event' columns."

    feature_names = sensitive_features.columns

    one_hot_encoded = pd.get_dummies(
        sensitive_features[feature_names], 
    )
    df_encoded = endpoint.join(one_hot_encoded)

    cph = CoxPHFitter(penalizer=0.0001)
    cph.fit(df_encoded, duration_col='time', event_col='event')  

    p_value = cph.log_likelihood_ratio_test().p_value          
    
    if p_value < 0.05 and (print_report or save_report):
        output = []

        print(f"⚠️  **Possible Bias Detected in {feature_names}** ⚠️")
        output.append(f"=== Subgroup Analysis for '{feature_names}' Using Cox Proportional Hazards Model ===\n")

        output.append("Model Statistics:")
        output.append(f"    AIC (Partial):               {cph.AIC_partial_:.2f}")
        output.append(f"    Log-Likelihood:              {cph.log_likelihood_:.2f}")
        output.append(f"    Log-Likelihood Ratio p-value: {p_value:.4f}")
        output.append(f"    Concordance Index (C-index):   {cph.concordance_index_:.2f}")

        summary_df = pd.DataFrame({
            'Feature': cph.summary.index.to_list(),
            'Coefficient': cph.summary['coef'].to_list(),
            'Standard Error': cph.summary['se(coef)'].to_list()
        })
        table_output = tabulate(summary_df, headers='keys', tablefmt='grid', showindex=False, floatfmt=".3f")
        output.append("Model Coefficients:")
        output.append('\n'.join(['    ' + line for line in table_output.split('\n')]))

        output_text = '\n'.join(output)
        
        if print_report:
            print(output_text)

        if save_report:
            with open(output_dir / f'{'_'.join(feature_names)}_CoxPH_summary.txt', 'w') as f:
                f.write(output_text)
    
    return cph, p_value

def intersectional_analysis(
        sensitive_features: pd.DataFrame,
        endpoint: pd.DataFrame,
        output_dir: Path | str,
        show_figure: bool = False,
        save_figure: bool = True, 
        tag: str = '',
        analysis_func: Callable = subgroup_analysis_OLS,
    ) -> None:
    """
    Performs intersectional analysis on `sensitive_features` and generates a p-value heatmap.

    This function evaluates pairwise interactions between sensitive features using 
    a specified analysis function (default: OLS). A heatmap of p-values is generated 
    to visualize statistical significance, with custom colormap boundaries and color gradients 
    to represent different p-value thresholds.

    Args:
        sensitive_features (pd.DataFrame): DataFrame containing categorical sensitive features.
        endpoint (pd.DataFrame): Target variable or survival data (depends on `analysis_func`).
        output_dir (Path | str): Directory to save the heatmap.
        show_figure (bool, optional): Whether to display the heatmap. Defaults to True.
        save_figure (bool, optional): Whether to display the heatmap. Defaults to False.
        tag (str, optional): Optional tag for filename customization. Defaults to ''.
        analysis_func (Callable, optional): Function to compute p-values (default: `subgroup_analysis_OLS`).

    Raises:
        AssertionError: If fewer than two features are provided.

    Returns:
        None
    """
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    feature_names = sensitive_features.columns
    assert len(feature_names) > 1, "This requires more than one sensitive feature"

    feat_pairs = list(combinations(feature_names, 2))
    pval_matrix = pd.DataFrame(index=feature_names, columns=feature_names, dtype=float)
    pval_list = []

    for col1, col2 in feat_pairs:
        _, pval = analysis_func(
            sensitive_features=sensitive_features[[col1, col2]], 
            endpoint=endpoint, 
            output_dir=output_dir, 
            print_report=False, 
            save_report=False
        )
        pval_list.append(pval)

    _, pvals_corrected, _, _ = multipletests(pval_list, method='fdr_bh')
    for (col1, col2), pval_corr in zip(feat_pairs, pvals_corrected):
        pval_matrix.loc[col1, col2] = pval_corr
        pval_matrix.loc[col2, col1] = pval_corr 

    pval_matrix = pval_matrix.apply(pd.to_numeric)
    mask = np.triu(np.ones_like(pval_matrix, dtype=bool))  # Keep only lower triangle
    np.fill_diagonal(mask, False)

    # Define custom gradient colormap
    colors = [
        (0.5, 0.0, 0.0),   # Dark Red (p ~ 0)
        (1.0, 0.0, 0.0),   # Red (p < 0.05)
        (1.0, 1.0, 0.0),   # Yellow (p ~ 0.05)
        (1.0, 0.5, 0.0),   # Orange (p ~ 0.1)
        (0.5, 0.5, 0.5)    # Gray (p > 0.1)
    ]
    cmap = mcolors.LinearSegmentedColormap.from_list("custom_pval", colors, N=256)

    # Normalize p-values with custom boundaries
    norm = mcolors.BoundaryNorm([0, 0.025, 0.05, 0.075, 0.1, 1], cmap.N)

    plt.figure(figsize=(10, 8))
    sns.heatmap(pval_matrix, annot=True, cmap=cmap, norm=norm, fmt=".3f", linewidths=0.5, cbar=True, mask=mask)

    title_tag = f" ({tag})" if tag else ""
    filename_tag = f"_{tag}" if tag else ""

    plt.title(f"P-value Heatmap OLS{title_tag}", fontsize=14, weight="bold")
    plt.xticks(rotation=45, ha="right")
    plt.yticks(rotation=0)
    plt.tight_layout()

    if save_figure:
        plt.savefig(output_dir / f"pval_heatmap_OLS{filename_tag}.png")
    
    if show_figure:
        plt.show()
    else:
        plt.close()



import pandas as pd

def format_leaderboard(leaderboard: pd.DataFrame, extra_metrics: list, score_col_name: str) -> pd.DataFrame:

    if score_col_name == 'score_val' and 'score_val' in leaderboard.columns:
        leaderboard = leaderboard.drop(score_col_name, axis=1)
    leaderboard = leaderboard.rename(columns={'score_test': score_col_name})

    def format_scores(row, score_col, extra_metrics):
        """Format scores as a string with AUROC, F1, and AUPRC. Or with R2 and RMSE for regression."""
        if 'f1' in extra_metrics:
            return f"AUROC {row[score_col]}\nF1: {row['f1']}\nAUPRC: {row['auprc']}"
        else:
            return f"R2 {row[score_col]}\nRMSE: {row['root_mean_squared_error']}"

    leaderboard[score_col_name] = leaderboard.apply(
        lambda row: format_scores(row, score_col_name, extra_metrics),
        axis=1
    )
    return leaderboard[['model', score_col_name]]

def aggregate_folds(consolidated_leaderboard:list, extra_metrics: list) -> pd.DataFrame:
    extra_metrics = [str(item) for item in extra_metrics]

    to_agg = {k: ['mean', 'min', 'max'] for k in ['score_test', *extra_metrics]}

    aggregated_leaderboard = consolidated_leaderboard.groupby('model').agg(to_agg).reset_index()

    final_leaderboard = pd.DataFrame({'model': aggregated_leaderboard['model']})

    for col in to_agg.keys():
        final_leaderboard[col] = [
            f'{round(row[0], 2)} [{round(row[1], 2)}, {round(row[2], 2)}]'
            for row in aggregated_leaderboard[col].values
        ]

    return final_leaderboard



from typing import List, Tuple

import pandas as pd
from pandas.api.types import is_numeric_dtype
from sklearn.impute import KNNImputer
from sklearn.preprocessing import MinMaxScaler

def replace_missing(
        data: pd.DataFrame, 
        categorical_columns: list,
        continuous_columns: list, config: dict
    ) -> pd.DataFrame:
    """
    Replace missing values in the dataset based on specified strategies.

    This method handles missing values in both continuous and categorical columns.
    The strategies for handling missing values are specified in the configuration.

    Continuous columns:
        - 'median': Replace missing values with the median of the column.
        - 'mean': Replace missing values with the mean of the column.
        - 'mode': Replace missing values with the mode of the column.
        - If an unknown strategy is specified, the median is used by default.

    Categorical columns:
        - A specified string from the configuration is used to replace missing values.
        - If 'knn' is specified, KNN imputation is used to fill missing values.

    Note:
        KNN imputation for categorical variables is performed using integer encoding and
        inverse transformation. This can be computationally expensive for large datasets.
    """
    for cont in continuous_columns:
        strategy = config['missingness_strategy']['continuous'].get(cont, 'median')
        if strategy.lower() == 'median':
            replace_with = data[cont].median()
        elif strategy.lower() == 'mean':
            replace_with = data[cont].mean()
        elif strategy.lower() == 'mode':
            replace_with = data[cont].mode()
        else:
            print(f'Unknown value {strategy} provided to replace {cont}. Using median')
            replace_with = data[cont].median()
        data[cont] = data[cont].fillna(replace_with)

    # For categorical var, replaces with string provided in config(default=Unknown)
    for cat in categorical_columns:
        filler = config['missingness_strategy']['categorical'].get(cat, 'Unknown')
        if config['missingness_strategy']['categorical'][cat].lower() != 'knn':
            # Adds filler as a new category
            data[cat] = data[cat].astype(str).fillna(filler).astype('category')

    data_to_use = data[categorical_columns + continuous_columns]
    for cat in categorical_columns:
        if config['missingness_strategy']['categorical'][cat].lower() == 'knn':
            print(f'Using KNN to fill missing values in {cat}, this may take a while...\n')
            data[cat] = knn_impute_categorical(data_to_use, categorical_columns)[cat]

    return data

def knn_impute_categorical(data: pd.DataFrame, categorical_columns: list) -> pd.DataFrame:
    """
    Perform KNN imputation on categorical columns.

    From https://www.kaggle.com/discussions/questions-and-answers/153147

    Args:
        data (DataFrame): The data containing categorical columns.
        categorical_columns (list): List of categorical column names.

    Returns
    -------
        DataFrame: Data with imputed categorical columns.
    """
    mm = MinMaxScaler()
    mappin = {}

    def find_category_mappings(df, variable):
        return {k: i for i, k in enumerate(df[variable].dropna().unique(), 0)}

    def integer_encode(df, variable, ordinal_mapping):
        df[variable] = df[variable].map(ordinal_mapping)

    df = data.copy()
    for variable in categorical_columns:
        mappings = find_category_mappings(df, variable)
        mappin[variable] = mappings

    for variable in categorical_columns:
        integer_encode(df, variable, mappin[variable])

    scaled_data = mm.fit_transform(df)
    knn_imputer = KNNImputer()
    knn_imputed = knn_imputer.fit_transform(scaled_data)
    df.iloc[:, :] = mm.inverse_transform(knn_imputed)
    for col in df.categorical_columns:
        df[col] = round(df[col]).astype('int')

    for col in categorical_columns:
        inv_map = {v: k for k, v in mappin[col].items()}
        df[col] = df[col].map(inv_map)

    return df

def get_outliers(data: pd.DataFrame, categorical_columns: list) -> Tuple[str, dict]:
    """
    Perform outlier analysis on categorical columns.

    Args:
        data (DataFrame): The data containing categorical columns.
        categorical_columns (list): List of categorical column names.

    Returns
    -------
        str: String of outliers
        dict: mapping for those outliers
    """
    mapping = {}
    total_report = ''


    for cat in categorical_columns:
        category_counts = data[cat].value_counts()
        threshold = int(len(data)*.01)
        outliers = category_counts[category_counts < threshold].index.tolist()

        mapping[cat] = {}

        for _cat in data[cat].unique():
            if _cat in outliers:
                mapping[cat][f'{_cat}'] = 'Other'
            else:
                mapping[cat][f'{_cat}'] = f'{_cat}'

        if len(outliers) > 0:
            outliers = [f'{o}: {category_counts[o]} out of {data[cat].count()}' for o in outliers]
            total_report += f'  - Outliers found in {cat}: {outliers}\n'
        else:
            total_report += f'  - No Outliers found in {cat}\n'

    return total_report, mapping

def infer_types(data: pd.DataFrame) -> Tuple[List[str], List[str], List[str]]:
    """
    Infer and categorize column data types in the dataset.

    Adapted from https://github.com/tompollard/tableone/blob/main/tableone/preprocessors.py

    This method analyzes the dataset to categorize columns as either
    continuous or categorical based on their data types and unique value proportions.

    Assumptions:
        - All non-numerical and non-date columns are considered categorical.
        - Boolean columns are not considered numerical but categorical.
        - Numerical columns with a unique value proportion below a threshold are
          considered categorical.

    The method also applies a heuristic to detect and classify ID columns
    as categorical if they have a low proportion of unique values.
    """
    date_columns = [
        col for col in data.select_dtypes(include=['object']).columns
        if pd.to_datetime(data[col], format='mixed', errors='coerce').notna().any()
    ]

    # assume all non-numerical and date columns are categorical
    numeric_cols = {col for col in data.columns if is_numeric_dtype(data[col])}
    numeric_cols = {col for col in numeric_cols if data[col].dtype != bool}
    likely_cat = set(data.columns) - numeric_cols
    likely_cat = list(likely_cat - set(date_columns))

    # check proportion of unique values if numerical
    for var in numeric_cols:
        likely_flag = 1.0 * data[var].nunique()/data[var].count() < 0.025
        if likely_flag:
            likely_cat.append(var)

    # Heuristic targeted at detecting ID columns
    likely_cat = [cat for cat in likely_cat if data[cat].nunique()/data[cat].count() < 0.2]

    categorical_columns = likely_cat
    continuous_columns = list(set(data.columns) - set(likely_cat) - set(date_columns))

    return categorical_columns, continuous_columns, date_columns

